; CHIP-F8
; -------
; CHIP-8 interpreter for F8
; on the Fairchild Channel F
; made by Tobias V. Langhoff
; edits by Fredric e5frog Blaholtz

	processor f8

;===================;
; Assembly Switches ;
;===================;

switch_music	=	0					; turns music on and off

;=========;
; Equates ;
;=========;

;------------;
; BIOS Calls ;
;------------;
clrscrn         =       $00d0					; uses r31
delay           =       $008f
pushk           =       $0107					; used to allow more subroutine stack space
popk            =       $011e
drawchar        =       $0679

;-------------------;
; Color Definitions ;
;-------------------;
red		=	$40
blue		=	$80
green		=	$00
bkg		=	$C0
clear		=	$FF

;-----------------;
; RAM Definitions ;
;-----------------;
ram			=	$2800					; use Schach RAM to hold CHIP-8 program
screenbuffer 	=	$2f00					; - " -          for screen buffer
chip8offset		=	$2600					; First $1FF is reserved, not using that. 

;--------------------;
; Keypad Definitions ;
;--------------------;

key0	=	%01000000	; L UP

; Use uninverted data for right controller
key1	=	%11101111	; R CCW
key2	=	%11110111	; R FWD
key3	= 	%11011111	; R CW
key4	=	%11111101	; R LEFT
key5	=	%01111111	; R DOWN
key6	=	%11111110	; R RIGHT
key7	=	%10111111	; R UP
key8	=	%11111011	; R BACK

key9	=	%00000001	; L RIGHT
keyA	=	%00000010	; L LEFT
keyB	=	%00000100	; L BACK
keyC	=	%00001000	; L FWD
keyD	=	%00010000	; L CCW
keyE	=	%00100000	; L CW
keyF	= 	%10000000	; L DOWN

;--------------------;
; Gfx setup          ;
;--------------------;

offsetX	=	22	; Center image
offsetY	=	12

;--------------------;
; Register Reference ;
;--------------------;

; The registers in the f8 are used as both RAM and registers for
; this game. The reference is as follows:

; r0: scratch
; r1-r2: current opcode
; r3: current palette
; r4-r9: scratch
; r10-r11 (H): Chip-8 PC
; r16-r31: CHIP-8 variables V0-VF
; r40-r41 (o50): I
; r60-63

; 020 r16 $10	V0
; 021 r17 $11	V1
; 022 r18 $12	V2
; 023 r19 $13	V3
; 024 r20 $14	V4
; 025 r21 $15	V5
; 026 r22 $16	V6
; 027 r23 $17	V7

; 030 r24 $18	V8
; 031 r25 $19	V9
; 032 r26 $1A	VA
; 033 r27 $1B	VB
; 034 r28 $1C	VC
; 035 r29 $1D	VD
; 036 r30 $1E	VE
; 037 r31 $1F	VF

; 040 r32 $20	I upper byte
; 041 r33 $21	I lower byte
; 042 r34 $22	** Used for sprite byte 1
; 043 r35 $23	Used for sprites ** Used for sprite byte 2
; 044 r36 $24	Used for sprites ** Timer
; 045 r37 $25	** Sound timer
; 046 r38 $26	** Pseudo random number
; 047 r39 $27	Stack pointer

; ***** TODO - make stack in RAM to fit 16 places.

; 050 r40 $28	Stack top
; 051 r41 $29
; 052 r42 $2A
; 053 r43 $2B
; 054 r44 $2C
; 055 r45 $2D
; 056 r46 $2E
; 057 r47 $2F

; 060 r48 $30
; 061 r49 $31
; 062 r50 $32
; 063 r51 $33
; 064 r52 $34
; 065 r53 $35
; 066 r54 $36
; 067 r55 $37

; 070 r56 $38
; 071 r57 $39
; 072 r58 $3A
; 073 r59 $3B
; 074 r60 $3C
; 075 r61 $3D	Sound timer
; 076 r62 $3E	Timer
; 077 r63 $3F	Pseudo random number


;===================;
; Main Program Code ;
;===================;

	MAC os
	; adds missing instruction: OS r, OR with a register
	; modifies r0
	com					; Invert A
	lr	0, A				; Save A in r0 (scratch)
	lr	A, {1}			; Copy from register in argument {1} to A
	com					; Invert this register value
	ns 0					; AND the inverted original and inverted new value, save in A
	com					; Invert the resulting value in A
						; Result in A, inverted original value in r0
	ENDM

;---------------;
; Program Entry ;
;---------------;

	org	$800

cartridgeStart:
	.byte	$55, $2B			; cartridge header

cartridgeEntry:
	lis	0				; init the h/w
	outs	1
	outs	4
	outs	5
	outs	0
                
	clr					; set to one color, black background
	lr	3, A				; clear entire Channel F screen to black
	pi	clrscrn			; use the ROM routine


copyGameToRAM:
	dci	gameROM
	xdc
	dci	ram
.copyByte:
	xdc
	lm					; copy byte from gameROM to A and advance DC0
	xdc					; swap DC
	st					; copy byte from A to gameRAM and advance DC0
	lr	Q, DC
	lr	A, QU				; check counter
	xi	$2f				; Reached $2F00?
	bnz	.copyByte

initInterpreter:
	dci	ram				; point at start of Chip-8 ROM copy (Chip-8 $200)
	lr	H, DC				; store Chip-8 PC (DC) in H (r10/r11)
	
	lisu	4				; set stack pointer located in o47
	lisl	7
	li	40				; Point to r40 (o50)
	lr	S, A



fetchDecodeLoop subroutine

	; decrease timer and sound delay
	lisu	7
	lisl	6
	clr
	as	S
	bz	timer_0
	ds	D				; Decrease timer if it's not 0, ISAR--

timer_0:
	clr
	as	S
	bz	timers_0
	ds	S

timers_0:

	lr	DC, H				; load Chip-8 PC from H
	lm					; fetch [PC] into A and advance
	lr	1, A				; store first opcode byte in r1
	lm					; fetch PC into A and advance
	lr	2, A				; store second opcode byte in r2
	lr	H, DC				; update PC


	; Two bytes fetched from Chip-8 code
	; Jump to proper part for the first digit.

	dci	.jumpTable
	lr	A, 1
	sr	4
	sl	1				; Upper to low nibble, double it
	adc					; Add to data counter
	lm
	lr	QU, A
	lm
	lr	QL, A
	lr	P0, Q				; Jump to jump table address
	
.jumpTable:
	.word firstDigitZero
	.word firstDigitOne
	.word firstDigitTwo
	.word firstDigitThree
	.word firstDigitFour
	.word firstDigitFive
	.word firstDigitSix
	.word firstDigitSeven
	.word firstDigitEight
	.word firstDigitNine
	.word firstDigitA
	.word firstDigitB
	.word firstDigitC
	.word firstDigitD
	.word firstDigitE
	.word firstDigitF



firstDigitZero:
	lr	A, 2
	xi	$E0				; $00E0 is clear screen
	bz	.clearScreen
	lr	A, 2
	xi	$EE				; $00EE is return from subroutine
	bz	.returnFromSubroutine
	jmp	fetchDecodeLoop		; Not checking for $0NNN call, fetch next word.


.clearScreen:
	; Clear the screenbuffer of 2048 pixels as 16x16 bytes, use r0
	lis	$F
	lr	0, A
	dci	screenbuffer
.clearMore
	clr
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	st
	ds	0
	bp	.clearMore


	li	32						; Rows in r0
	lr	0, A
screenClear:
	lr	A, 0
	ai	offsetY
	com							; invert it
.screenClearRow:
	outs	5						; load into port 5 (row) latch

	; load the width into the horizontal counter
	li	64
	lr	1, A

	clr							; load the VRAM x coordinate
	ai	offsetX
	com							; invert it
.screenClearColumn:
	outs	4						; load into port 4 (column)
.screenClearDrawBit:
	; Set color
	lis	$C
	sl	4
.screenClearSavePixel:
	outs	1						; output A in p1 (color)

.screenClearTransferData:
	; transfer the pixel data
	lis	6
	sl	4						; %01100000  $60
	outs	0
	sl	1						; %11000000  $C0
	outs	0

	lis	2						; Graphics update delay
screenClear.savePixelDelay:
	ai	$ff
	bnz	screenClear.savePixelDelay


.screenClearCheckColumn:
	ds	1						; decrease r1 (horizontal counter)
	bz	.screenClearNewRow			; if it's 0 it's done

	ins	4						; get p4 (column)
	ai	$ff						; add 1 (complemented)
	br	.screenClearColumn			; branch

.screenClearNewRow
	ds	0
	bnz	screenClear

.screenClearExit:

	; Done, go for next opcode
	jmp	fetchDecodeLoop






.returnFromSubroutine:
	lisu	4				; Stack pointer
	lisl	7
	lr	A, S				; Fetch the pointer, put in ISAR
	lr	IS, A

;	lr	A, i				; *** If o57 is increased it will become o50, there's NO carry.
	lr	A, S
	lr	HU, A

	lr	A, IS
	inc
	lr	IS, A				; Increase ISAR with carry

	lr	A, S
	lr	HL, A

	lisu	4
	lisl	7
	ds	s				; decrease stack back by two *** ERROR CHECK? ***
	ds	s
	
	jmp	fetchDecodeLoop



firstDigitOne:
						; $1NNN jump to address NNN
	lr	A, 1				; load first byte of opcode
	sl	4				; remove first nibble, just keep *N**
	sr	4
	ai	$26				; add RAM offset (first $200 not used)
	lr	HU, A				; load into PC (H register)
	lr	A, 2				; load second byte of opcode
	lr	HL, A				; load into PC (H register)
	jmp	fetchDecodeLoop




firstDigitTwo:				; Call subroutine at NNN
	lisu	4
	lisl	7
	lr	A, S				; load stack pointer in A
	inc					; increase stack pointer by two
	inc
	lr	S, A				; Store as new stack pointer
	lr	IS, A				; load new stack pointer into ISAR

	lr	A, HU				; push PC onto stack
	lr	S, A				; Store Hu in S
	
	lr	A, IS				; Increase ISAR when - at any position
	inc
	lr	IS, A

	lr	A, HL
	lr	S, A				; Return address stored

	lr	A, 1				; load first byte of opcode
	sl	4				; remove first nibble
	sr	4
	ai	$26				; add RAM offset
	lr	HU, A				; load into PC
	lr	A, 2				; load second byte of opcode
	lr	HL, A				; load into PC
	jmp	fetchDecodeLoop




firstDigitThree:				; 3XNN, if (Vx == NN) skip next instrucion

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $3x..
	lr	IS, A

	lr	A, 2
	xs	s
	bnz	.notEqual

	lr	DC, H				; load PC
	lm
	lm
	lr	H, DC				; DC two bytes forward, store in H.
.notEqual
	jmp	fetchDecodeLoop




firstDigitFour:				; 4XNN, if (Vx != NN) skip next instrucion

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, 2
	xs	s
	bz	.equal

	lr	DC, H				; load PC
	lm
	lm
	lr	H, DC				; Skipped two bytes
.equal:
	jmp	fetchDecodeLoop
	



firstDigitFive subroutine		; 5XY0, if (Vx == Vy) skip next instruction

	; Check if opcode is invalid?
	; Last digit should be 0, is it proper to skip to next?

;	lr	A, 2
;	sl	4
;	bz	.notEqual

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $5X
	lr	IS, A

	lr	A, S
	lr	0, A

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S				; Fetch Vy
	xs	0				; Compare them
	bnz	.notEqual

	lr	DC, H				; load PC
	lm
	lm
	lr	H, DC
.notEqual:
	jmp	fetchDecodeLoop




firstDigitSix:				; 6XNN, set Vx = NN

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A				; V-register set

	lr	A, 2				; get second byte of opcode
	lr	S, A				; set it
	jmp	fetchDecodeLoop




firstDigitSeven:				; 7XNN, add NN to Vx, no carry, Vx += NN

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $7x..
	lr	IS, A

	lr	A, S				; get current value of VX
	as	2				; add value of second byte of opcode
	lr	S, A				; set new value of VX
	jmp	fetchDecodeLoop




firstDigitEight subroutine		; 8XY* Vx is set

	dci	.jumpTable
	lr	A, 2				; get second byte of opcode
	sl	4
	sr	4				; remove first nibble for type of operation
	sl	1				; Double to add it twice
	adc
	lm
	lr	QU, A
	lm
	lr	QL, A
	lr	P0, Q				; Store read address in P0 - jump
	
.jumpTable:
	.word .lastDigitZero
	.word .lastDigitOne
	.word .lastDigitTwo
	.word .lastDigitThree
	.word .lastDigitFour
	.word .lastDigitFive
	.word .lastDigitSix
	.word .lastDigitSeven

	.word	fetchDecodeLoop		; 8 not used
	.word	fetchDecodeLoop		; 9 not used
	.word	fetchDecodeLoop		; A not used
	.word	fetchDecodeLoop		; B not used
	.word	fetchDecodeLoop		; C not used
	.word	fetchDecodeLoop		; D not used

	.word .lastDigitE	

	.word	fetchDecodeLoop		; F not used




.lastDigitZero:				; 8XY0, Vx = Vy

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A				; Proper ISAR for Vy

	lr	A, S
	lr	0, A				; Store value from Vy in r0

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, 0
	lr	S, A				; Store in Vx

	jmp	fetchDecodeLoop




.lastDigitOne:				; 8XY1, Vx = Vx (bitwise OR) Vy

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	lr	2, A				; Store value from Vy in r2

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, S				; Fetch value from Vx

	os	2				; OR with register 2 (uses MACRO, don't use r0)
	lr	S, A				; Store in Vx


	; QUIRK, Octo sets Vf to 0
	; Set VF
	lisu	3
	lisl	7
	clr
	lr	S, A

	jmp	fetchDecodeLoop




.lastDigitTwo:				; 8XY2, Vx = Vx (bitwise AND) Vy

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	lr	0, A				; Store Vy value in r0

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, S
	ns	0
	lr	S, A				; Result into Vx


	; QUIRK, Octo sets Vf to 0
	; Set VF
	lisu	3
	lisl	7
	clr
	lr	S, A


	jmp	fetchDecodeLoop




.lastDigitThree:				; 8XY3, Vx = Vx XOR Vy


	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	lr	0, A

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, S				; Fetch value from Vx

	xs	0
	lr	S, A				; EXOR with r0 and result in Vx


	; QUIRK, Octo sets Vf to 0
	; Set VF
	lisu	3
	lisl	7
	clr
	lr	S, A

	jmp	fetchDecodeLoop




.lastDigitFour:				; 8XY4, Vx += Vy, add Vy to Vx, Carry on Vf

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	lr	0, A				; Value in Vy in r0


	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, S
	as	0
	lr	S, A				; Add values

	; Set VF
	lisu	3
	lisl	7

	lr	J, W				; Copy status register to J (J is r9)
	lr	A, J
	ni	2				; Keep b1
	sr	1				; Move to b0
	lr	S, A				; Store carry in Vf

	jmp	fetchDecodeLoop




.lastDigitFive:				; 8XY5, Vx=Vx-Vy, subtract Vy from Vx, If Vx >= Vy then Vf=1 else 0


	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	com
	inc
	lr	0, A				; Number in Vy negative and stored in r0

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, S
	as	0
	lr	S, A				; Add Vx and -Vy.

	; Set VF
	lisu	3
	lisl	7

	lr	J, W				; Copy status register to J (r9)
	lr	A, J
	ni	2
	sr	1
	lr	S, A				; Borrow same as carry?

	jmp	fetchDecodeLoop




.lastDigitSix:				; 8XY6, Vx >>= 1, Store LSB in VF, shift Vx right 1

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A
	lr	0, A				; Store ISAR# in r0

	; Check b0
	lr	A, S
	ni	%00000001			; Keep b0
	lisu	3
	lisl	7
	lr	S, A				; Store b0 in Vf

	lr	A, 0				; Fetch Vx ISAR
	lr	IS, A

	lr	A, S				; Get value
	sr	1				; Shift right
	lr	S, A				; Save back shifted value

	jmp	fetchDecodeLoop




.lastDigitSeven:				; 8XY7, Vx = Vy - Vx, Vf set to 0 when borrow


	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A
	lr	1, A				; Store ISAR value

	lr	A, S				; Make Vx value negative
	com
	inc
	lr	0, A



	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	as	0
	lr	0, A

	; Set VF
	lisu	3
	lisl	7

	lr	J, W
	lr	A, J
	ni	2
	sr	1
	lr	S, A

	; returns ISAR pointing at VX
	lr	A, 1
	lr	IS, A

	lr	A, 0
	lr	S, A

	jmp	fetchDecodeLoop




.lastDigitE:				; 8XYE, Vx <<= 1, MSB in Vf, shift Vx left 1

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A
	lr	0, A				; Store ISAR in r0

	lr	A, S
	; Check b7
	sr	4
	sr	1
	sr	1
	sr	1				; Move b7 to b0
	lisu	3
	lisl	7
	lr	S, A				; Store in Vf

	lr	A, 0				; Fetch Vx ISAR
	lr	IS, A

	lr	A, S				; Get value
	sl	1				; Shift left
	lr	S, A				; Save back shifted value

	jmp	fetchDecodeLoop

	


firstDigitNine subroutine			; 9XY0, skip if (Vx != Vy), skip next if not equal

	; TODO: handle if last digit is not 0?

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111
	lr	IS, A
	lr	A, S				; Fetch and save in r0
	lr	0, A

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble
	oi	$10
	lr	IS, A

	lr	A, S
	xs	0				; XOR values

	bz	.equal

	lr	DC, H				; load PC
	lm
	lm
	lr	H, DC				; Skips past an instruction
.equal
	jmp	fetchDecodeLoop




firstDigitA:				; ANNN, sets I = NNN
	lisu	4
	lisl	0

	lr	A, 1				; load first byte of opcode
	sl	4				; remove first nibble
	sr	4
	ai	$26				; add RAM offset

	lr	I, A				; load into I-register

	lr	A, 2				; load second byte of opcode
	lr	S, A				; load into I

	jmp	fetchDecodeLoop
	



firstDigitB:				; BNNN, PC = V0 + NNN

	lr	A, 1				; load first byte of opcode
	sl	4				; clear first nibble
	sr	4
	ai	$26				; add RAM offset

	lr	QU, A				; load into Q
	
	lr	A, 2				; load second byte of opcode
	lr	QL, A				; load into PQ

	lr	DC, Q				; Set DC (Chip-8 PC)

	lisu	2
	lisu	0
	lr	A, S
	adc

firstDigitB_noAdd:
	lr	H, DC				; load into PC
	jmp	fetchDecodeLoop

	


firstDigitC:				; Vx = rand() AND NN, bitwise AND between random and NN

; In this example the pseudo random number is stored in register o77
; Also uses register 0 for temporary storage

	pi	LFSR

	lr	A, 2				; Second OP-code, NN
	ns	S				; AND with the "random" number
	lr	0, A				; temp store in r0

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	lr	A, 0
	lr	S, A				; Store bitwise AND in Vx


	jmp	fetchDecodeLoop


LFSR:             ; Linear Feedback Shift Register
	lisu	7
	lisl	7
	clr
	as	S
	bz	doEor
	lr	0, A				; Save original number in r0
	sl	1				; shift one step left
	lr	S, A
	bz	noEor

	; Do XOR if b7 was 1, workaround for missing carry on "sl" 
	lr	A, 0
	ni	%10000000
	bz	noEor
doEor:
	lr	A, S
	xi	$2b     			; Do the XOR thing
	lr	S, A
noEor:
	pop




firstDigitD:				; DXYN, Draw(Vx, Vy, N), Object 8xN at coordinates
						; Bit coded gfx from memory location I
						; Vf = 1 if any pixel changes from set to unset
						; XOR draw at coordinate (Vx, Vy). 
						; One byte wide, max 15 bytes high


	; returns ISAR pointing at VX
	lr	A, 1				; Dx = %1101.... keep b4
	ni	%00011111			; What we have left: $1X
	lr	IS, A				; Vx is in ISAR $10-$1F (starts at o20)

	lr	A, S				; get X coordinate from Vx
	ni	$07				; bit offset for sprite data [0,7]
	lr	7, A				; Keep bit offset in scratch 7

	lr	A, S				; get X again
	ni	$3F				; Strip to modulo 64 (max width)

	sr	1				; Which Byte does it start in?
	sr	1
	sr	1
	lr	8, A				; store buffer Byte position X in scratch 8

	lr	A, 2				; get second byte of opcode
	sl	4				; clear first nibble
	sr	4
	lr	9, A				; save in r9 as display row counter (N)

	lr	A, 2
	sr	4				; Move upper nibble to lower nibble (get Y coordinate)
	oi	$10
	lr	IS, A

	lr	A, S				; get Y value
	ni	$1F				; modulo 32 (mask off max height)

	; r2 is free to use, store Y coordinate for blit
	lr	2, A

	sl	1
	sl	1				; Multiply by 8 (8 bytes per row) 
	sl	1				; To find the row in buffer

	as	8				; sum "X byte" and "Y byte" to get to the byte in the buffer

	lr	QL, A				; Set the pixel position in buffer [$00,$FF]
	li	$2f				; the screenbuffer address
	lr	QU, A
	lr	DC, Q				; Load buffer position in DC0

	xdc					; Swap DC counters


	; Position in RAM where to plot the sprite is now in DC1

	; load I into DC via Q, address location for graphics, I untouched
	lisu	4
	lisl	0				; I-register
	lr	A, I
	lr	QU, A
	lr	A, S
	lr	QL, A

	lr	DC, Q

	; Address where to fetch r9 row bytes is now in DC0

	; clear VF before calculating graphics
	lisu	3
	lisl	7
	clr
	lr	S, A

; r0 
; r1 X for blit8
; r2 Y for blit8
; r3 current bit count
; r4 graphics byte to plot 1
; r5 graphics byte to plot 2
; r6 blit8 horizontal counter
; r7 bit offset
; r8 
; r9 row counter
; DC0 sprite data start address
; DC1 screen buffer with offset
; ISAR setup for VF

.nextSpriteRow:
	clr					; Clear the two sprite registers
	lr	4, A
	lr	5, A

	as	9				; Check r9 if rows are completed (set flags)
	bz	.displaySpriteExit	; If 0 we're done, exit.
.dontResetI:
	ds	9				; decrease row counter
	lm					; get one byte of sprite data, set next address
	lr	4, A				; byte in r4
	lr	A, 7				; get bit offset and copy to r3 to use as counter
	lr	3, A

.splitSpriteRow:
	lr	A, 3				; get current bit count
	ns	3				; AND a with itself to set flags (zero?)
	bz	.storeSpriteRow		; if 0 sprite data is ready
	lr	A, 4				; load the sprite from r4
	ni	1				; Keep b0 then select carry or not carry
	bz	.rightShiftWithNoCarry
	
.rightShiftWithCarry:
	lr	A, 4				; Sprite data from r4
	sr	1				; shift it right once
	lr	4, A				; Store shifted data
	lr	A, 5				; Load next sprite byte
	sr	1				; shift right once
	oi	$80				; Add the shifted bit
	lr	5, A				; load shifted byte back to r5
	ds	3				; Decrease bit count
	jmp	.splitSpriteRow

.rightShiftWithNoCarry:
	lr	A, 4				; Sprite data from r4
	sr	1				; shift it right once
	lr	4, A				; Store shifted byte in r5
	lr	A, 5				; Next byte from r5
	sr	1				; Shift right
	lr	5, A				; Store back in r5
	ds	3
	jmp	.splitSpriteRow

.storeSpriteRow:
	xdc					; switch to screen buffer pointer
	lr	Q, DC				; Backup DC in Q so we can revert here
	lr	A, 4				; First assembled byte in r4
	nm					; and with screen buffer to detect collision
	bz	.noCollision1
	lis	1
	lr	S, A				; Collision, store in Vf
.noCollision1:
	lr	DC, Q				; Restore screen buffer DC
	lr	A, 4				; First assembled byte in r4
	xm					; xor with screen buffer (this increments dc)
	lr	DC, Q				; restore dc
	st					; store xor-ed result in same spot, DC0 increases


	lr	4, A				; Store EXOR:ed result in r4 for plot
	; Set X-coordinate, one of 8x [0,7]
	lr	A, QL
	ni	%00000111
	sl	1
	sl	1
	sl	1
	lr	1, A
	pi	blit8


	; check for screen boundary, we don't wrap sprites
	lr	A, QL
	lr	3, A
	lr	Q, DC				; store new DC in Q so we can revert here
	lr	A, QL
	xs	3				; xor old and new QL
	ni	$08				; see if we crossed from $2F?7 to $2F?8; if so, we've wrapped around X
	lis	7				; go to next row in screen data
	bnz	.outOfBounds		; Don't plot the second byte

	lr	A, 5				; Fetch byte 2
	nm					; and with screen data, check for byte 2 collision
	bz	.noCollision2
	lis	1
	lr	S, A				; Set Vf to change (collision)
.noCollision2:
	lr	DC, Q				; restore dc

	lr	A, 5				; Fetch byte 2
	xm					; xor with screen data (this increments dc)
	lr	DC, Q				; restore dc
	st					; store xor-ed result DC+

	lr	4, A				; Copy data to r4 for blit8
	lis	8
	as	1
	lr	1, A				; Update X coordinate to next byte in the row
	pi	blit8

	lis	6				; go to next row in screen data

.outOfBounds:
	adc					; Now at next row
	xdc					; swap back to I pointer for next sprite row

	lis	1				; Increase Y-coordinate for next row
	as	2
	lr	2, A

	jmp	.nextSpriteRow

.displaySpriteExit:


	jmp	fetchDecodeLoop







	
firstDigitE:				; Skip next instruction if Vx key is/is not pressed
	lr	A, 2
	xi	$9e
	bnz	.ex9e
	lr	A, 2
	xi	$a1
	bnz	.exa1
	jmp	fetchDecodeLoop

.ex9e:					; Skips the next instruction if the key stored in VX is pressed. 
						; if (key() == Vx)

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4				; Fastest lower nibble mask %0000xxxx
	oi	$10				; %0001xxxx, registers for Vx
	lr	IS, A


	; Fetch right controller data
	clr
	outs	0
	outs	1				; Allow RHC, clear port
	ins	1				; Read port
	; Not inverting it
	lr	4, A
	com
	bnz	.ex9e_indata


	; Fetch left controller data
	clr
	outs	4				; Clear port
	ins	4				; Read port
	com					; Invert data
	lr	4, A
	bz	.ex9e_exit			; No key was pressed, exit.

.ex9e_indata:
	; Compare data
	dci	.ex9e_hcdata
	lr	A, S				; Fetch key# from Vx
	adc					; Find corresponding HC indata in ROM
	lr	A, 4				; Fetch read HC indata
	xm					; EXOR indata and memory (same results in zero)
	bz	.exa1_skip			; If zero, key and controller data is the same, skip.

	; It was not active, exit and don't skip next op
.ex9e_exit:
	jmp	fetchDecodeLoop




.exa1:					; Skips the next instruction if the key stored in VX is not pressed.
						; if (key() != Vx)

	; returns ISAR pointing at VX
	lr	A, 1
	sl	4
	sr	4				; Fastest lower nibble mask %0000xxxx
	oi	$10				; %0001xxxx, registers for Vx
	lr	IS, A


	; Fetch right controller data
	clr
	outs	0
	outs	1				; Allow RHC, clear port
	ins	1				; Read port
	; Not inverting it
	lr	4, A
	com
	bnz	.exa1_indata


	; Fetch left controller data
	clr
	outs	4				; Clear port
	ins	4				; Read port
	com					; Invert data
	bz	.exa1_skip			; No key at all was pressed, exit with skip.
	lr	4, A


	; Key is pressed... is it the one in Vx?
.exa1_indata:
	; Compare data
	dci	.ex9e_hcdata
	lr	A, S				; Fetch key# from Vx
	adc					; Find corresponding HC indata in ROM
	lr	A, 4				; Fetch read HC indata
	xm					; EXOR indata and memory (same results in zero)
	bnz	.exa1_skip			; Not the same key - skip.

	; Same key - continue program. 
.exa1_exit:
	jmp	fetchDecodeLoop


.exa1_skip:
	lr	DC, H
	lm
	lm
	lr	H, DC
	jmp	fetchDecodeLoop


	
firstDigitF:
	; TODO convert below to jump table (or branch/offset table if space permits)

	lr	A, 2				; get second byte of opcode
	xi	$07
	bnz	.not07
	jmp	.lastNibble07
.not07:
	lr	A, 2
	xi	$0A
	bnz	.not0A
	jmp	.lastNibble0A
.not0A:
	lr	A, 2
	xi	$15
	bnz	.not15
	jmp	.lastNibble15
.not15:
	lr	A, 2
	xi	$18
	bnz	.not18
	jmp	.lastNibble18
.not18:
	lr	A, 2
	xi	$1E
	bnz	.not1E
	jmp	.lastNibble1E
.not1E:
	lr	A, 2
	xi	$29
	bnz	.not29
	jmp	.lastNibble29
.not29:
	lr	A, 2
	xi	$33
	bnz	.not33
	jmp	.lastNibble33
.not33:
	lr	A, 2
	xi	$55
	bnz	.not55
	jmp	.lastNibble55
.not55:
	lr	A, 2
	xi	$65
	bnz	.not65
	jmp	.lastNibble65
.not65:	
	jmp	fetchDecodeLoop

.lastNibble07:				; Sets VX to the value of the delay timer.
						; Vx = get_delay()

	lisu	7
	lisl	6
	lr	A, S
	lr	0, A				; Fetch timer value, save in r0	


	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx..
	lr	IS, A

	lr	A, 0
	lr	S, A				; Store timer value in Vx


	jmp	fetchDecodeLoop


.lastNibble0A:				; A key press is awaited, and then stored in VX. Halted until next key event
						; Vx = get_key()

	; Wait for any key, store which one in Vx

	; Set ISAR to point at VX
	lr	A, 1
	sl	4
	sr	4
	oi	$10
	lr	IS, A

	clr
	lr	 S, A				; Store 0 in Vx


.lastNibble0A_repeat:

	; Right controller
	clr
	outs	0
	outs	1
	ins	1
	lr	4, A				; Backup uninverted right HC data
	com
	bnz	.lastNibble0A_key		; Exit if there's a keypress

	; Left controller
	clr
	outs	4
	ins	4
	com
	bz	.lastNibble0A_repeat
	lr	4, A

.lastNibble0A_key:

	dci	.ex9e_hcdata		; Set data counter to position for key values.

.lastNibble0A_count:
	ds	S				; Decrease before comparing (first turns to $FF)
	lr	A, 4				; Fetch pressed key
	xm					; Compare to ROM
	bnz	.lastNibble0A_count	; Repeat if it's not the same

	; Inverted result is in Vx
	lr	A, S
	com
	lr	S, A

	; Done.
	jmp	fetchDecodeLoop
	

.lastNibble15:				; Sets the delay timer to VX.
						; delay_timer(Vx)
	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx..
	lr	IS, A

	lr	A, S				; Fetch Vx

	lisu	7
	lisl	6
	lr	S, A				; Store from Vx in timer

	jmp	fetchDecodeLoop

.lastNibble18:				; Sets the sound timer to VX.
						; sound_timer(Vx)

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx..
	lr	IS, A

	lr	A, S				; Fetch Vx value

	lisu	7
	lisl	5
	lr	S, A				; Store from Vx in sound_timer


	jmp	fetchDecodeLoop

.lastNibble1E:				; Adds VX to I-register. VF is not affected
						; I += Vx
	lisu	4
	lisl	0

	lr	A, I
	lr	QU, A
	lr	A, S
	lr	QL, A

	lr	DC, Q				; Set DC from I-register


	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx..
	lr	IS, A

	; Now we want to add the value in the register selected by the ISAR to the address in DC.
	; But ADC treats the value in A as a two's complement number, so we need to do this in a roundabout way:

	lr	A, S				; load VX offset into A
	ns	s				; AND the offset with itself to check the upper bit
	bp	.adcHack			; if bit 7 is not set, we just proceed as normal
	sr	1				; if bit 7 is set, shift right
	adc					; add this offset to DC
	adc					; twice
	lr	A, S				; load the original offset again
	ni	1				; get bit 0 so we can finally add that if needed
.adcHack:
	adc					; add offset

	lisu	4				; set ISAR to I
	lisl	0

	lr	Q, DC				; load DC back into Q
	lr	A, QU
	lr	I, A				; Store DC in I-register
	lr	A, QL
	lr	S, A

	jmp	fetchDecodeLoop

.lastNibble29:				; Sets I to the location of the sprite for the character in VX. 
						; Characters 0-F (in hexadecimal) are represented by a 4x5 font.
						; I = sprite_addr[Vx]

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx.. => $1x
	lr	IS, A

	lr	A, S				; Fetch value from VX
	sl	1				; multiply by 4
	sl	1
	as	S				; Also add the number: we multiply by 5

	dci	font				; set DC to the fontset's address
	adc					; add the offset for the current character

	lr	Q, DC				; Store value of data counter into Q

	lisu	4				; set ISAR to I-register
	lisl	0

	lr	A, QU				; load Q into I-register via A
	lr	I, A
	lr	A, QL
	lr	S, A				; I-register is now loaded with data address for font

	jmp	fetchDecodeLoop

.lastNibble33:				; Stores the binary-coded decimal representation of VX
						; Place the hundreds digit in memory at location in I
						; Tens digit at location I+1 and the ones digit at location I+2
						; set_BCD(Vx) ;*(I+0) = BCD(3);  *(I+1) = BCD(2); *(I+2) = BCD(1);

	; returns ISAR pointing at VX
	lr	A, 1
	ni	%00011111			; Keep b4 from $Fx..
	lr	IS, A

	; BCD routine courtesy of https://my.eng.utah.edu/~nmcdonal/Tutorials/BCDTutorial/BCDConversion.html

	lis	0
;	lr	0, A				; bit counter
	lr	1, A				; hundreds (treat as nibble)
	lr	2, A				; tens (treat as nibble)
	lr	3, A				; ones (treat as nibble)

	lr	A, S
	lr	4, A				; r4: the binary number to convert to BCD

	lis	8
	lr	0, A

.loop:
	lr	A, 1				; if hundreds > 4 then add 3
	ci	4
	bp	.testTens
	ai	3
	lr	1, A
.testTens:					; if tens > 4 then add 3
	lr	A, 2
	ci	4
	bp	.testOnes
	ai	3
	lr	2, A
.testOnes:					; if ones > 4 then add 3
	lr	A, 3
	ci	4
	bp	.shift
	ai	3
	lr	3, A

.shift:
	lr	A, 1				; shift hundreds nibble left
	sl	1
	sl	4
	sr	4
	lr	1, A

	lr	A, 2
	ni	$08				; if tens most significant bit is 1, carry it over to hundreds
	bz	.shiftTens
	lr	A, 1
	inc
	lr	1, A
.shiftTens:					; shift tens nibble left
	lr	A, 2
	sl	1
	sl	4
	sr	4
	lr	2, A

	lr	A, 3
	ni	$08				; if ones most significant bit is 1, carry it over to tens
	bz	.shiftOnes
	lr	A, 2
	inc
	lr	2, A
.shiftOnes:					; shift ones nibble left
	lr	A, 3
	sl	1
	sl	4
	sr	4
	lr	3, A

	lr	A, 4
	ni	$80				; if binary number's most significant bit is 1, carry it over to ones
	bz	.shiftByte
	lr	A, 3
	inc
	lr	3, A
.shiftByte:					; shift binary number left
	lr	A, 4
	sl	1
	lr	4, A

	ds	0					; decrement bit counter
	bnz	.loop				; if we haven't processed the entire binary number, continue

	lisu	4
	lisl	0

	lr	A, I
	lr	QU, A
	lr	A, S
	lr	QL, A

	lr	DC, Q				; Set DC to value in I

	lr	A, 1
	st					; store hundreds in address I
	lr	A, 2
	st					; store tens in address I+1
	lr	A, 3
	st					; store ones in address I+3

	jmp	fetchDecodeLoop



.lastNibble55:				; FX55
						; Stores V0 to VX (including VX) in memory starting at address I. 
						; The offset from I is increased by 1 for each value written
						; I itself is left unmodified.
						; reg_dump(Vx, &I)
	lisu	4
	lisl	0

	lr	A, I
	lr	QU, A
	lr	A, S
	lr	QL, A

	lr	DC, Q				; DC is set to the memory address

	lr	A, 1
	sl	4
	sr	4
	lr	0, A

	lisu	2
	lisl	0

.storeLoop:
	lr	A, i
	st
	ds 0
	bp .storeLoop			; if positive number, loop

	; TODO quirk, store DC back in I via Q

	jmp	fetchDecodeLoop

.lastNibble65:				; Fills V0 to VX (including VX) with values from memory starting at address I.
						; The offset from I is increased by 1 for each value written.
						; I is left unmodified.
						; reg_load(Vx, &I)
	lisu	4
	lisl	0				; Fetch I

	lr	A, I
	lr	QU, A
	lr	A, S
	lr	QL, A

	lr	DC, Q				; Fetch address from I, store in data counter

	lr	A, 1
	sl	4
	sr	4
	lr	0, A				; Mask out X and store in r0

	lisu	2				; Start putting data in V0
	lisl	0

.loadLoop:
	lm					; Read data at I address, DC automatically increased
	lr	S, A				; Store data in V-register
	lr	A, IS
	inc
	lr	IS, A				; Increase ISAR (next V-register)

	ds	0				; Decrease X (amount of registers to change)
	bp	.loadLoop			; if positive number, loop

	; TODO quirk, store DC back in I via Q
	jmp	fetchDecodeLoop




.ex9e_hcdata:
	; Data for HC corresponding to each key 0-F
	.byte key0,key1,key2,key3,key4,key5,key6,key7
	.byte key8,key9,keyA,keyB,keyC,keyD,keyE,keyF


	; Font data should be present at Chip-8 offset + 0x050-0x0A0
font:
	; 0
	.byte	%11110000
	.byte	%10010000
	.byte	%10010000
	.byte	%10010000
	.byte	%11110000

	; 1
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000

	; 2
	.byte	%11110000
	.byte	%00010000
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000

	; 3
	.byte	%11110000
	.byte	%00010000
	.byte	%11110000
	.byte	%00010000
	.byte	%11110000

	; 4
	.byte	%10010000
	.byte	%10010000
	.byte	%11110000
	.byte	%00010000
	.byte	%00010000

	; 5
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000
	.byte	%00010000
	.byte	%11110000

	; 6
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000
	.byte	%10010000
	.byte	%11110000

	; 7
	.byte	%11110000
	.byte	%00010000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000

	; 8
	.byte	%11110000
	.byte	%10010000
	.byte	%11110000
	.byte	%10010000
	.byte	%11110000

	; 9
	.byte	%11110000
	.byte	%10010000
	.byte	%11110000
	.byte	%00010000
	.byte	%11110000

	; A
	.byte	%11110000
	.byte	%10010000
	.byte	%11110000
	.byte	%10010000
	.byte	%10010000

	; B
	.byte	%11100000
	.byte	%10010000
	.byte	%11100000
	.byte	%10010000
	.byte	%11100000

	; C
	.byte	%11110000
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%11110000

	; D
	.byte	%11100000
	.byte	%10010000
	.byte	%10010000
	.byte	%10010000
	.byte	%11100000

	; E
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000

	; F
	.byte	%11110000
	.byte	%10000000
	.byte	%11110000
	.byte	%10000000
	.byte	%10000000



	org $0e00


	include	"blit8.asm"



; The CHIP-8 program will be copied from ROM to Schach RAM, of which
; we have $700 or 1792 available bytes (the final $FF or 256 bytes are
; reserved for the screen buffer). We assert both the size of the
; Channel F ROM and the CHIP-8 ROM like so:

	.byte "    GAMECODE    "

gameROM:
;	incbin "test_opcode.ch8"
	incbin "pong.ch8"
;	incbin "helloworld.ch8"
;	incbin "KeyTest.ch8"
;	incbin "TIMEBOMB.ch8"
;	incbin "X-MIRROR.ch8"
;	incbin "15 Puzzle [Roger Ivie].ch8"
;	incbin "Tank.ch8"
;	incbin "IBM.ch8"

	IF	0

OPCODE TEST:

3XNN	00EE	8XY5
4XNN	8XY0	8XY6
5XY0	8XY1	8XYE
7XNN	8XY2	FX55
9XY0	8XY3	FX33
ANNN	8XY4	1NNN




Keypad Test, by hap, 15-02-06

press a chip8 key and the pressed char will light up
if you want to do something funny, soft-reset the chip8/emulator over and over,
and the sprite layout will become messed up ;p

chip8 keypad:
1 2 3 c
4 5 6 d
7 8 9 e
a 0 b f


00000200   12 4E   	Goto $24E
00000202   08 19   	DATA
00000204   01 01   
00000206   08 01   
00000208   0F 01   
0000020A   01 09   
0000020C   08 09   
0000020E   0F 09   
00000210   01 11   
00000212   08 11   
00000214   0F 11   
00000216   01 19   
00000218   0F 19   
0000021A   16 01   
0000021C   16 09   
0000021E   16 11   
00000220   16 19   
00000222   FC FC   
00000224   FC FC   
00000226   FC FC   
00000228   FC 00   
0000022A   A2 02   	Sub: I=$202
0000022C   82 0E   		MSB in Vf, Left shift V2
0000022E   F2 1E   		Add V2 to I
00000230   82 06   		LSB in Vf, Right shift V2
00000232   F1 65   		Fill V0 to V1 from memory in I ($202)
00000234   00 EE   	Return
00000236   A2 02   	Sub: I=$202
00000238   82 0E   		MSB in Vf, Left shift V2
0000023A   F2 1E   		Add V2 to I
0000023C   82 06   		LSB in Vf, Right shift V2
0000023E   F1 55   		Store V0 to V1 at memory address in I ($202)
00000240   00 EE   	Return
00000242   6F 10   	Sub: Set Vf to $10
00000244   FF 15   		Set delay timer to Vf - which is $10
00000246   FF 07   		Set Vf to the value of delay timer
00000248   3F 00   		If Vf=$00 skip next
0000024A   12 46   		Jump to $246 (Wait for timer to timeout)
0000024C   00 EE   	Return
0000024E   00 E0   	Clear screen
00000250   62 00   	Set V2 to $00
00000252   22 2A   	Call Sub $22A
00000254   F2 29   	Sprite location in I for number in V2
00000256   D0 15   	draw(V0,V1,5) Draw sprite at (V0,V1) five rows
00000258   70 FF   	Add $FF to V0 (decrease V0)
0000025A   71 FF   	Add $FF to V1 (decrease V1)
0000025C   22 36   	Call Sub $236
0000025E   72 01   	Add $01 to V2
00000260   32 10   	Skip next if V2 = $10
00000262   12 52   	Jump to $252, repeat drawing all characters
00000264   F2 0A   	Wait for keypress, store it in V2
00000266   22 2A   	Call Sub $22A
00000268   A2 22   	I=$222
0000026A   D0 17   	draw(V0,V1,7) Draw sprite at (V0,V1) seven rows
0000026C   22 42   	Call Sub $242
0000026E   D0 17   	draw(V0,V1,7) Draw sprite at (V0,V1) seven rows
00000270   12 64   	Jump to $264 (which waits for keypress)


	ENDIF




cartridgeEnd:
	org gameROM + $6ff,0
	.byte $00
